/* 
** DAV connection, request and response process
** Copyright (C) 2001 SungHun Kim <hunkim@kebi.com>
**
** This library and program is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**   
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
** MA 02111-1307, USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libdav.h>
#include <nsocket.h>
#include <base64.h>
#include <dates.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>

#include <config.h>

/* procsee DEBUG option */
#ifdef SHOW_REQ
#define SHOW_REQUEST(x)  printf("%s\n",((x)))
#else
#define SHOW_REQUEST(x)  
#endif

/* Free if ptr has value */
void safe_free(void *ptr) {
   if(ptr) 
     free(ptr);
 
    return;
}

/* check null before return */
char *safe_strdup(const char *s) {
    char *ptr = strdup(s);
    
    if(!ptr) {
	/* complain about the lack of memory */
	fprintf(stderr, "STRDUP: ");
	perror(s);
	return ptr;
    }
    
    return ptr;
}

/* check null before return */
char *safe_strndup(const char *s, size_t len) {
    char *ptr = malloc(len+1);

    if(!ptr) {
	/* complain about the lack of memory */
	fprintf(stderr, "STRNDUP: ");
	perror(s);
	return ptr;
    }

    strncpy(ptr, s, len);
    ptr[len] = 0;
    
    return ptr;
}


/* clear record */
void dav_rec_clear(dav_rec *rec) {
    memset(rec, 0x00, sizeof(dav_rec));
    
    return;
}

/* feee all thing in rec */
void dav_rec_free(dav_rec *rec) {
    safe_free(rec->hh.allow);
    safe_free(rec->hh.location);
    safe_free(rec->hh.lock_token);
    safe_free(rec->hh.content_type);
    
    safe_free(rec->dest);
    safe_free(rec->xml);
    safe_free(rec->lock_token);

    /* any thing else? */
    
    return; 
}

/*
** parse head
** read return code
** keep-alive
** content-lengh
** transfer-encoding
*/
static int parse_head(dav_config *dc, char *str, http_head *hh) {
    char *delm=" \n\r\t";
    char *ptr;

    switch(*str) {
    case 'H':
    case 'h':               /*123456789*/
	if(!strncasecmp(str, "HTTP/", 5)) {
	    ptr = strtok(str+8, delm);
	    if(ptr)
	      hh->ret_code = atoi(ptr);
	}
	break;

    case 'C':
    case 'c': 
	if(!strncasecmp(str, "Content-Length: ",16)) {
	    hh->has_content_length = 1;
	    hh->content_length = atoi(str+16);
                                  /*12345678901234567890123*/
	} else if(!strncasecmp(str, "Connection: Keep-Alive",23)) {
	   dc->connected = 1;
        }
	   
	break;

    case 'A':
    case 'a':               /*1234567*/ 
	if(!strncasecmp(str, "Allow: ",7)) {
	    hh->allow = safe_strdup(str+7);
	}

	break;
    case 'T':
    case 't':               /*1234567890123456789*/ 
	if(!strncasecmp(str, "Transfer-Encoding: ",19)) {
	                           /*1234567*/
	    if(!strncasecmp(str+19, "chunked",7))
		hh->chunked = 1;
	}
    case 'L':
    case 'l':               /*1234567890*/ 
	if(!strncasecmp(str, "Location: ",10)) {
	    hh->location = safe_strdup(str+10);
	                           /*123456789012*/
	} else if(!strncasecmp(str, "Lock-Token: ",12)) {
	    ptr = strtok(str+12, delm);
	    if(ptr)
		hh->lock_token = safe_strdup(ptr);
	}
	break;
    }
    return 1;
}

/* print out the body */
int dav_printf(dav_config *dc, dav_rec *rec) {
    int i, n, tn = 0;    
    int to_read;
    char buf[BUF_SIZE];

    for(i=0;tn < rec->hh.content_length;i++, tn+=n) { 
	to_read = min(rec->hh.content_length-tn, BUF_SIZE-1);
	if((n=sock_read(dc->ns, buf, to_read)) <0 ) 
	    break;
	
#ifdef SHOW_BODY
	printf(buf);
#endif
    }
    return 0;
}

/* dave the body into out fp */
int dav_save(dav_config *dc, dav_rec *rec) {
    int i, n, tn = 0;    
    int to_read;
    char buf[BUF_SIZE];
    
    lseek(rec->u.out, 0, SEEK_SET);
    
    /* no content-length and closed */
    if(rec->u.out<1 || rec->hh.ret_code!=200) {
	if(rec->hh.content_length) {
	    /* just read and throw */
	    to_read = min(rec->hh.content_length-tn, BUF_SIZE-1);
	    if((n=sock_read(dc->ns, buf, to_read)) <0 ) 
		goto out;
	}
    } else if(!dc->connected && !rec->hh.has_content_length && !rec->hh.chunked) {
	while((n=read(dc->sock, buf, sizeof(buf)-1)) > 0) 
	    write(rec->u.out, buf, n);
    } else {
	for(i=0;tn < rec->hh.content_length;i++, tn+=n) { 
	    to_read = min(rec->hh.content_length-tn, BUF_SIZE-1);
	    if((n=sock_read(dc->ns, buf, to_read)) <0 ) 
		goto out;
	    
	    write(rec->u.out, buf, n);
	    
	}
    }
    
 out:    
    if(rec->u.out<1) {
	dc->errorno = NULLFILE;
	return -1;
    }
    
    return tn;
}

static int read_head(dav_config *dc, dav_rec *rec) {
    char buf[BUF_SIZE];
    int n;
    int tn = 0;

again:
    /* set clear */
    memset(&rec->hh, 0x00, sizeof(rec->hh));

    while((n=sock_readline(dc->ns, buf, BUF_SIZE-1)) > 0) {
	tn+=n;
#ifdef SHOW_HEAD
	printf("RES: %s", buf);
#endif
	/* reset connect flag */
        dc->connected = 0;
	parse_head(dc, buf, &rec->hh);
	if(*buf=='\n' || *buf=='\r') break;
	
    }

    /* 100 continue */
    if(rec->hh.ret_code==100)
        goto again;

#ifdef SHOW_HEAD
    printf("DAV_RET: %d\n", rec->hh.ret_code);
#endif

    return tn;
}

/* read chunked data */
static int read_chunk(dav_config *dc, dav_rec *rec, 
		      int (*do_func)(dav_config *, dav_rec *)) {
    int chunk_size;
    int tn = 0;
    char buf[BUF_SIZE];

    while(1) {
	/* read chunk size */
	if(sock_readline(dc->ns, buf, BUF_SIZE-1) < 0) 
	    break;

#ifdef SHOW_HEAD
	printf("Chunk-size : %s\n", buf);
#endif
	tn += chunk_size = strtol(buf, NULL, 16);
	/* 0, return */
	if(!chunk_size) 
	    break;
	    
	/* set content_length */
	rec->hh.content_length = chunk_size;
	
	/* do something */
	do_func(dc, rec);

	/* read CRLF */
	if(sock_readline(dc->ns, buf, BUF_SIZE-1) < 0)
	    break;
    } 
    return 0;
}

/* read body with function pointer */
static int read_body(dav_config *dc, dav_rec *rec, 
	      int (*start_func)(dav_config *, dav_rec *), 
	      int (*do_func)(dav_config *, dav_rec *),
	      int (*end_func)(dav_config *, dav_rec *))
{
    int tn;

    /* rec function has priority */
    if(rec->start_func) start_func = rec->start_func;
    if(rec->do_func)    do_func    = rec->do_func;
    if(rec->end_func)   end_func   = rec->end_func;

    if(start_func) 
	start_func(dc, rec);

    /* do func should read rec->hh.content_leng */
    if(!do_func)
	do_func = dav_printf;

    /* is chunked data ? */
    if(rec->hh.chunked) 
	tn = read_chunk(dc, rec, do_func);
    else 
	tn = do_func(dc, rec);
    
    if(end_func) 
	end_func(dc, rec);
    
    return tn;
}

/*
** send basic request include method, host, useragent
*/
static int start_http_request(dav_config *dc, char *method) {
    char buf[BUF_SIZE];
    int tn=0;

    if(!dc->connected) {
#ifdef DEBUG
       puts("RECONNECT");
#endif
        sock_close(dc->ns);
	dc->ns = dav_reopen(dc);
    }
    
    if(dc->proxy_host)
        sprintf(buf, "%s %s://%s:%d%s HTTP/1.1", method, 
		dc->scheme, dc->host, dc->port,  dc->path);
    else 
        sprintf(buf, "%s %s HTTP/1.1", method, dc->path);
    
    tn+=sock_sendline(dc->ns, buf);
    SHOW_REQUEST(buf);

    sprintf(buf, "Host: %s:%d", dc->host, dc->port);
    tn+=sock_sendline(dc->ns, buf);
    SHOW_REQUEST(buf);

    tn+=sock_sendline(dc->ns, "User-Agent: DAV-FS/0.0.1");
    
    tn+=sock_sendline(dc->ns, "Connection: Keep-Alive");
    
    /* if we need id and passwd */
    if(dc->auth_type && dc->authorization) {
	sprintf(buf, "Authorization: %s %s", 
	       dc->auth_type, dc->authorization);
	tn+=sock_sendline(dc->ns, buf);
	SHOW_REQUEST(buf);
    }

    return tn;
}


/*
** send your head like as content Length
*/
static int send_http_request(dav_config *dc, char *name, char *value) {
    char buf[BUF_SIZE];
    sprintf(buf, "%s: %s", name, value);
    SHOW_REQUEST(buf);
    return sock_sendline(dc->ns, buf);
}


/*
** send your head like as content Length using integer
*/
static int send_http_request_int(dav_config *dc, char *name, int value) {
    char buf[BUF_SIZE];
    sprintf(buf, "%s: %d", name, value);
    SHOW_REQUEST(buf);    
    return sock_sendline(dc->ns, buf);
}
  
/*
** send CRLF
*/  
static int end_http_request(dav_config *dc) {
    return sock_sendline(dc->ns, "");
}

/* parse URL 
** 
** This is part of cadver 0.17.0
**
*/
int dav_parseurl(dav_config *dc, char *uri) {
    const char *pnt, *slash, *colon;
    
    dc->port = -1; 
    dc->host = NULL; 
    dc->path = NULL; 
    dc->scheme = NULL;
    
    pnt = strstr(uri, "://");
    if (pnt) {
        dc->scheme = safe_strndup(uri, pnt - uri);
        pnt += 3; /* start of hostport segment */
    } else {
        pnt = uri;
    }
    
    slash = strchr(pnt, '/');
    colon = strchr(pnt, ':');
    if (slash == NULL) {
        dc->path = safe_strdup("/");
        if (colon == NULL) {
            dc->host = safe_strdup(pnt);
        } else {
            dc->port = atoi(colon+1);
            dc->host = safe_strndup(pnt, colon - pnt);
        }
    } else {
        if (colon == NULL || colon > slash) {
            /* No port segment */
            dc->host = safe_strndup(pnt, slash - pnt);
        } else {
            /* Port segment */
            dc->port = atoi(colon + 1);
            dc->host = safe_strndup(pnt, colon - pnt);
        }
        dc->path = safe_strdup(slash);
    }

    return 0;
}

/* parse Proxy URL 
** 
** This is part of cadver 0.17.0
**
*/
static int dav_parseproxy(dav_config *dc, char *proxy) {
    const char *pnt, *slash, *colon;
        
    pnt = strstr(proxy, "://");
    if (pnt) {
        dc->proxy_scheme = safe_strndup(proxy, pnt - proxy);
        pnt += 3; /* start of hostport segment */
    } else {
        pnt = proxy;
    }
    
    slash = strchr(pnt, '/');
    colon = strchr(pnt, ':');
    if (slash == NULL) {
	if (colon == NULL) {
            dc->proxy_host = safe_strdup(pnt);
        } else {
            dc->proxy_port = atoi(colon+1);
            dc->proxy_host = safe_strndup(pnt, colon - pnt);
        }
    } else {
        if (colon == NULL || colon > slash) {
            /* No port segment */
            dc->proxy_host = safe_strndup(pnt, slash - pnt);
        } else {
            /* Port segment */
            dc->proxy_port = atoi(colon + 1);
            dc->proxy_host = safe_strndup(pnt, colon - pnt);
        }
    }

    return 0;
}

/* reopen connection with url, id, pass information */
nsocket * dav_reopen(dav_config *dc) {
    struct in_addr addr;
    int ssl_mode = 0;

    /* if sock has a connection */
/*
    if(dc->ns)
	sock_close(dc->ns);
*/

    if(dc->proxy_host && dc->proxy_port) { 
	/* Support for proxy */
	if(sock_name_lookup(dc->proxy_host, &addr)==-1) {
	    return (dc->ns=NULL);
	}
	
	dc->ns = sock_connect(addr, dc->proxy_port, &(dc->sock));

	/* check SSL peoxy */
	if(dc->scheme && !strcmp(dc->scheme, "https"))
	    ssl_mode = 1;
    } else {
	if(sock_name_lookup(dc->host, &addr)==-1) {
	    return (dc->ns=NULL);
	}
	
	dc->ns = sock_connect(addr, dc->port, &(dc->sock));

	/* check SSL support */
	if(dc->scheme && !strcmp(dc->scheme, "https"))
	    ssl_mode = 1;
    }

    if(ssl_mode) {
#ifdef ENABLE_SSL    
        sock_make_secure(dc->ns, NULL);
#else
        perror("Not support SSL!");
	return NULL;
#endif
    }

    dc->connected = 1;
    return dc->ns;
}


/*
 * get some datas and fill dc structure 
 */
int dav_prepare(dav_config *dc, char *url, char *id, char *pass, char *proxy) {
    char buf[BUF_SIZE];

    if(dav_parseurl(dc, url) || dc->host==NULL) 
	return -1;
    
    /* not assigned */
    if(dc->port == -1) 
	dc->port = 
	  (dc->scheme && !strcmp(dc->scheme, "https"))?
	  DEFAULT_SSL_PORT:DEFAULT_PORT;  
    
    if(proxy) {
	dav_parseproxy(dc, proxy);
	if(dc->port == -1) 
	    dc->port = 
	      (dc->scheme && !strcmp(dc->scheme, "https"))?
	      DEFAULT_SSL_PORT:DEFAULT_PORT;  
    }

    /* Add autorization */
    if(id && pass) {
	char *base64;
	
	sprintf(buf,"%s:%s", id, pass);

	if((base64  = ne_base64(buf))==NULL) {
	    perror("base64");
	    exit(-2);
	}
	
	dc->authorization = safe_strdup(base64);
	dc->id = safe_strdup(id);
	dc->pass = safe_strdup(pass);
	dc->auth_type = safe_strdup("Basic");
	
	safe_free(base64);
    }

    return 0;
    
}

/* open connection with url, id, pass information */
nsocket * dav_open(dav_config *dc, char *url, char *id, char *pass, char *proxy) {
    if(dav_prepare(dc, url, id, pass, proxy)==-1)
	return NULL;

    return dav_reopen(dc);
}

/*
** free dc  
*/
int dav_dc_free(dav_config *dc) {
    safe_free(dc->scheme);
    safe_free(dc->host);
    safe_free(dc->path);
    safe_free(dc->auth_type);
    safe_free(dc->authorization);
    safe_free(dc->id);
    safe_free(dc->pass);
    
    /* free proxy things */
    safe_free(dc->proxy_host);
    safe_free(dc->proxy_scheme);

    /* any thing else? */

    return 0;
}


/* 
 *   Close only connection  
 *   should call dav_dc_free(dc) to free dc area
 */
int dav_only_close(dav_config *dc) {
    return sock_close(dc->ns);
}

/* 
 * dav_close and free dc memory 
 */
int dav_close(dav_config *dc) {
    dav_only_close(dc);
    return dav_dc_free(dc);
}

int dav_put(dav_config *dc, dav_rec *rec) {
    char lock[BUF_SIZE];
    /* send method and Host and agent basig stuff */
    start_http_request(dc, "PUT");
    send_http_request_int(dc, "Content-Length", rec->o.length);
    
    /* if locked resource */
    if(rec->lock_token) {
	sprintf(lock, "(%s)", rec->lock_token);
	send_http_request(dc, "If", lock);
    }
    
    end_http_request(dc);
    
    /* send_body should send rec->u.length size */
    if(rec->send_body) {
	rec->send_body(dc, rec);
    } 
   
    /* get response */
    read_head(dc, rec);
    
    read_body(dc, rec, NULL, dav_printf, NULL);
    
    return rec->hh.ret_code;
}    

 
int dav_get(dav_config *dc, dav_rec *rec)  {
   
    /* send method and Host and agent basig stuff */
    start_http_request(dc, "GET");
    end_http_request(dc);
   
    /* get response */
    read_head(dc, rec);
    
    read_body(dc, rec, NULL, dav_save, NULL);
    
    return rec->hh.ret_code;
}


/* 
** do dav get command and fill head information
**
** reading body data is program's (who call this) responsbility
**
*/ 
int dav_get_head(dav_config *dc, dav_rec *rec)  {
   
    /* send method and Host and agent basig stuff */
    start_http_request(dc, "GET");
    end_http_request(dc);
   
    /* get response */
    read_head(dc, rec);
    
    return rec->hh.ret_code;
}


int dav_lock ( dav_config *dc, dav_rec *rec ) {
    int length = rec->xml?strlen(rec->xml):0;

    start_http_request(dc, "LOCK");
    send_http_request_int(dc, "Content-Length",  length);
    send_http_request(dc, "Content-Type", "text/xml; charset=\"utf-8\"");
    send_http_request(dc, "Timeout", LOCK_TIMEOUT);
    end_http_request(dc);

    /*
    ** send xml
    */
    if(rec->xml) {
        sock_fullwrite(dc->ns, rec->xml, length);
#ifdef DEBUG
	puts(rec->xml);
#endif
    }
    
    /*
    ** get response
    */
    read_head(dc, rec);

    /* on multi response */
    if(rec->hh.ret_code==207 || rec->hh.ret_code==200 ) 
	read_body(dc, rec, NULL, dav_printf, NULL);

    return rec->hh.ret_code;
}    

/*
** unlock
** should have a velue in rec->lock_token
**
*/
int dav_unlock ( dav_config *dc, dav_rec *rec ) {
    start_http_request(dc, "UNLOCK");
    if(rec->lock_token)
	send_http_request(dc, "Lock-Token", rec->lock_token);
    send_http_request_int(dc, "Content-Length",  0);
    send_http_request(dc, "Content-Type", "text/xml; charset=\"utf-8\"");
    end_http_request(dc);

    /*
    ** get response
    */
    read_head(dc, rec);
    read_body(dc, rec, NULL, dav_printf, NULL);
    return rec->hh.ret_code;
}    

int dav_propfind ( dav_config *dc, dav_rec *rec ) {
    int length = rec->xml?strlen(rec->xml):0;

    start_http_request(dc, "PROPFIND");
    send_http_request_int(dc, "Depth", rec->o.depth);
    send_http_request_int(dc, "Content-Length",  length);
    send_http_request(dc, "Content-Type", "text/xml; charset=\"utf-8\"");
    end_http_request(dc);

    /*
    ** send xml
    */
    if(rec->xml) {
        sock_fullwrite(dc->ns, rec->xml, length);
#ifdef DEBUG
	puts(rec->xml);
#endif
    }
    
    /*
    ** get response
    */
    read_head(dc, rec);

    /* on multi response */
    if(rec->hh.ret_code==207 || rec->hh.ret_code==200 ) 
	read_body(dc, rec, NULL, dav_printf, NULL);

    return rec->hh.ret_code;
}    

int dav_proppatch ( dav_config *dc, dav_rec *rec ) {
    
    /* no xml or length */
    if(!rec->xml || !rec->o.length) 
	return 0;
	
    start_http_request(dc, "PROPPATCH");
    send_http_request_int(dc, "Content-Length", rec->o.length);
    send_http_request(dc, "Content-Type", "text/xml");
    end_http_request(dc);
       
    /*
    ** send xml
    */
    sock_fullwrite(dc->ns, rec->xml, rec->o.length);
    
    /* get response */
    read_head(dc, rec);
    
    read_body(dc, rec, NULL, dav_printf, NULL);
   
    return rec->hh.ret_code;
}    


int dav_mkcol ( dav_config *dc, dav_rec *rec ) {
     
    start_http_request(dc, "MKCOL");
    end_http_request(dc);
    
    /* get response */
    read_head(dc, rec);
    
    read_body(dc, rec, NULL, dav_printf, NULL);
 
    return rec->hh.ret_code;
}    


int dav_delete ( dav_config *dc, dav_rec *rec ) {
    start_http_request(dc, "DELETE");
    end_http_request(dc);
   
    /* get response */
    read_head(dc, rec);
    
    read_body(dc, rec, NULL, dav_printf, NULL);
    
    return rec->hh.ret_code;;
    
}    


int dav_copy ( dav_config *dc, dav_rec *rec) {
    
    start_http_request(dc, "COPY");
    send_http_request(dc, "Destination", rec->dest);
    send_http_request(dc, "Overwrite", (rec->o.overwrite?"T":"F"));
    end_http_request(dc);
    
    /* get response */
    read_head(dc, rec);
    
    read_body(dc, rec, NULL, dav_printf, NULL);
     
    return rec->hh.ret_code;;
}    


int dav_move ( dav_config *dc, dav_rec *rec) {
    
    start_http_request(dc, "MOVE");
    
    send_http_request(dc, "Destination", rec->dest);
    send_http_request(dc, "Overwrite", (rec->o.overwrite?"T":"F"));
    end_http_request(dc);
    
    /* get response */
    read_head(dc, rec);
    
    read_body(dc, rec, NULL, dav_printf, NULL);
  
    return rec->hh.ret_code;;
}    


/*
** get options
*/
int dav_options ( dav_config *dc, dav_rec *rec) {
    
    start_http_request(dc, "OPTIONS");
    end_http_request(dc);
    
    /* get response */
    read_head(dc, rec);
    
    read_body(dc, rec, NULL, dav_printf, NULL);

    return rec->hh.ret_code;     
}    


/*
** head method
*/
int dav_head ( dav_config *dc, dav_rec *rec) {
    start_http_request(dc, "HEAD");
    end_http_request(dc);
    
    /* get response */
    read_head(dc, rec);
    
    return rec->hh.ret_code;     
}    

/* process date information */
time_t dav_parse_date(const char *date) {
    time_t tmp;
    if(!date) return -1;

    tmp = ne_rfc1123_parse(date);
    if (tmp == -1) {
        tmp = ne_rfc1036_parse(date);
        if (tmp == -1) {
            tmp = ne_asctime_parse(date);
	    if(tmp == -1) {
		tmp = iso8601_parse(date);
            }
	}
    }

#ifdef DEBUG
    printf("D: %s %d\n", date, (int)tmp);
#endif
    return tmp;
}

